"""
Copyright (c) 2001, MetaSlash Inc.  All rights reserved.

PyChecker is a tool for finding common bugs in python source code.
It finds problems that are typically caught by a compiler for less
dynamic languages, like C and C++.  It is also similar to lint.

Contact Info:
  http://pychecker.sourceforge.net/
  pychecker-list@lists.sourceforge.net
"""

# A version # to check against in the main module (checker.py)
# this will allow us to check if there are two versions of checker
# in site-packages and local dir
MAIN_MODULE_VERSION = 3

